use cgmath::{vec3, Vector3};

use image::ImageBuffer;
use utilities::prelude::*;

use std::f32::{MAX, MIN};

mod triangle;
use triangle::Triangle;

mod aabb;
use aabb::AABB;

mod ray;
use ray::Ray;

mod view;
use view::View;

mod camera;
use camera::Camera;

fn main() {
    let input_data = [
        Triangle::new(
            vec3(0.0, 0.0, 0.0),
            vec3(1.0, 0.0, 0.0),
            vec3(1.0, 1.0, 0.0),
        ),
        Triangle::new(
            vec3(1.0, 1.0, 0.0),
            vec3(0.0, 1.0, 0.0),
            vec3(0.0, 0.0, 0.0),
        ),
        Triangle::new(
            vec3(1.0, 1.0, 0.0),
            vec3(2.0, 1.0, 0.0),
            vec3(2.0, 2.0, 0.0),
        ),
        Triangle::new(
            vec3(2.0, 2.0, 0.0),
            vec3(1.0, 2.0, 0.0),
            vec3(1.0, 1.0, 0.0),
        ),
    ];

    let view = View {
        position: vec3(0.0, -4.0, 4.0),
        look_at: vec3(0.0, 0.0, 0.0),
        up: vec3(0.0, 0.0, 1.0),
        fov: 45.0,
    };

    let camera = Camera::new(
        vec3(0.0, -4.0, 4.0),
        vec3(0.0, 0.0, 0.0),
        vec3(0.0, 0.0, 1.0),
        45.0,
        1280.0 / 720.0,
    );

    let _debug_ray = camera.primary_ray(640, 360, 1280, 720);

    //debug_raytracer_view(1280, 720, &view, &input_data);
    debug_raytracer_camera(1280, 720, &camera, &input_data);
}

fn f_to_u(color: f32) -> u8 {
    ((color * 255.0) / 1.0) as u8
}

/*
fn calculate_ray(x: u32, y: u32, dim_x: u32, dim_y: u32, view: &View) -> Ray {
    let aspect_ratio = dim_x as f32 / dim_y as f32;

    let fov = view.fov / M_PI * 2.0 / 180.0;

    let p_x = (1.0 - 2.0 * ((x as f32 + 0.5) / dim_x as f32)) * fov * aspect_ratio;
    let p_y = (2.0 * (((y) as f32 + 0.5) / dim_y as f32) - 1.0) * fov;

    let look_at = Matrix4::look_at(
        Point3::from_vec(view.position),
        Point3::from_vec(view.look_at),
        vec3(0.0, 0.0, 1.0),
    );

    let test = look_at_test(view.position, view.look_at, vec3(0.0, 0.0, 1.0));

    let camera_to_world = look_at.invert().unwrap();

    let origin = camera_to_world * vec4(0.0, 0.0, 0.0, 1.0);
    let direction = camera_to_world * vec4(p_x, p_y, -1.0, 1.0);

    Ray::new(origin.truncate(), -direction.truncate().normalize())
}
*/

/*
fn look_at_test(eye: Vector3<f32>, center: Vector3<f32>, up: Vector3<f32>) -> Matrix4<f32> {
    let mut look_at_matrix = Matrix4::identity();

    let forward = (eye - center).normalize();
    let right = up.cross(forward).normalize();
    let up = forward.cross(right);

    look_at_matrix[0][0] = right.x;
    look_at_matrix[0][1] = right.y;
    look_at_matrix[0][2] = right.z;

    look_at_matrix[1][0] = up.x;
    look_at_matrix[1][1] = up.y;
    look_at_matrix[1][2] = up.z;

    look_at_matrix[2][0] = forward.x;
    look_at_matrix[2][1] = forward.y;
    look_at_matrix[2][2] = forward.z;

    look_at_matrix[3][0] = eye.x;
    look_at_matrix[3][1] = eye.y;
    look_at_matrix[3][2] = eye.z;

    look_at_matrix
}
*/

fn debug_raytracer_camera(dim_x: u32, dim_y: u32, camera: &Camera, data: &[Triangle]) {
    let mut imgbuf = ImageBuffer::new(dim_x, dim_y);

    let acceleration_data = create_acceleration_data(data);

    for (x, y, pixel) in imgbuf.enumerate_pixels_mut() {
        let ray = camera.primary_ray(x, y, dim_x, dim_y);

        let color = pixel_color(&ray, data, &acceleration_data);

        *pixel = image::Rgb([f_to_u(color.x), f_to_u(color.y), f_to_u(color.z)]);
    }

    imgbuf.save("raytrace.png").unwrap();
}

fn debug_raytracer_view(dim_x: u32, dim_y: u32, view: &View, data: &[Triangle]) {
    let mut imgbuf = ImageBuffer::new(dim_x, dim_y);

    let acceleration_data = create_acceleration_data(data);

    for (x, y, pixel) in imgbuf.enumerate_pixels_mut() {
        let ray = Ray::primary_ray(x as f32, y as f32, dim_x as f32, dim_y as f32, view);

        let color = pixel_color(&ray, data, &acceleration_data);

        *pixel = image::Rgb([f_to_u(color.x), f_to_u(color.y), f_to_u(color.z)]);
    }

    imgbuf.save("raytrace.png").unwrap();
}

fn create_acceleration_data(input_data: &[Triangle]) -> Vec<AABB> {
    let mut acceleration_data = Vec::new();

    let accel_count = (input_data.len() as f32 / 24.0).ceil() as usize;

    for i in 0..accel_count {
        let mut accel = AABB::new();

        let mut bbmin = vec3(MAX, MAX, MAX);
        let mut bbmax = vec3(MIN, MIN, MIN);

        // 3 vertices per triangles
        // 8 triangles per acceleration structure
        let start_index = 8 * i;
        let mut end_index = 8 * (i + 1);

        // check if end_index exceeds input data length
        if end_index > input_data.len() {
            end_index = input_data.len();
        }

        // set index information
        accel.start_index = start_index;
        accel.end_index = end_index;

        // create bounding box
        for k in start_index..end_index {
            for vertex in input_data[k].points.iter() {
                // check minimum values
                bbmin.x = bbmin.x.min(vertex.x);
                bbmin.y = bbmin.y.min(vertex.y);
                bbmin.z = bbmin.z.min(vertex.z);

                // check maximum values
                bbmax.x = bbmax.x.max(vertex.x);
                bbmax.y = bbmax.y.max(vertex.y);
                bbmax.z = bbmax.z.max(vertex.z);
            }
        }

        accel.bounds[0] = bbmin;
        accel.bounds[1] = bbmax;

        acceleration_data.push(accel);
    }

    acceleration_data
}

fn pixel_color(ray: &Ray, data: &[Triangle], acceleration_data: &Vec<AABB>) -> Vector3<f32> {
    let mut final_color = vec3(0.0, 1.0, 0.0);
    let mut closest_value = MAX;
    let mut closest_index = -1;

    for accel in acceleration_data.iter() {
        if accel.intersect(ray) {
            for i in accel.start_index..accel.end_index {
                let mut t = 0.0;

                if data[i].intersect_mt(ray, &mut t) {
                    if t < closest_value {
                        closest_index = i as i32;
                        closest_value = t;
                    }
                }
            }
        }
    }

    if closest_index != -1 {
        final_color = vec3(1.0, 0.0, 0.0);
    }

    return final_color;
}
